require 'observer'

class ObservableDelegate
  include Observable

  def initialize(subject, notify_with: nil, observe: [], delegate: [])
    @subject = subject
    @notify_with = notify_with
    @observe_methods = observe
    @delegate_methods = delegate
  end

  def method_missing(method, *args, &block)
    if should_delegate?(method) && should_notify?(method)
      result = @subject.send(method, *args, &block)

      changed

      payload = if @notify_with
        @subject.send(@notify_with)
      else
        result
      end

      notify_observers(payload)

      result
    elsif should_delegate?(method)
      @subject.send(method, *args, &block)
    else
      # create a no method error here and remove the
      # last backtrace from it so that it points to
      # where this method was used.
      begin
        raise NoMethodError.new("No method #{method} found for #{@subject.class}")
      rescue NoMethodError => e
        e.set_backtrace(e.backtrace[1..-1])
        raise e
      end
    end
  end

  # should delegate if the subject has the method
  # and the method is explicitly provided or there
  # are not methods explicitly provided
  def should_delegate?(method)
    @subject.methods.include?(method) &&
      (@delegate_methods.include?(method) ||
       @delegate_methods.empty?)
  end

  # should notify if the method is explicitly provied or
  # no methods are explicitly provided
  def should_notify?(method)
    @observe_methods.include?(method) || @observe_methods.empty?
  end
end
